/*
 * Copyright (c) 2014 by ELECTRONIC ASSEMBLY <technik@lcd-module.de>
 * EA DOG Graphic (ST7565R) software library for arduino.
 *
 * This file is free software; you can redistribute it and/or modify
 * it under the terms of either the GNU General Public License version 2
 * or the GNU Lesser General Public License version 2.1, both as
 * published by the Free Software Foundation.
 */

#include <Arduino.h>
#include <SPI.h>
#include <avr/pgmspace.h>

#include "dog.h"


#define INITLEN 13
byte init_DOGS102[INITLEN] = {0x40, 0xA1, 0xC0, 0xA4, 0xA6, 0xA2, 0x2F, 0x27, 0x81, 0x10, 0xFA, 0x90, 0xAF};

byte led_lookup[] = 
            {
              0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 15, 16, 17, 18, 19, 21,
              22, 23, 25, 26, 27, 29, 30, 32, 33, 35, 36, 38, 39, 41, 43, 44, 46, 48, 50, 52,
              53, 55, 57, 59, 61, 63, 65, 67, 70, 72, 74, 76, 78, 81, 83, 86, 88, 91, 93, 96,
              98,101,104,107,110,112,115,118,121,125,128,131,134,138,141,145,148,152,155,159,
              163,167,171,175,179,183,187,191,196,200,205,209,214,219,224,229,234,239,244,250,
              255 
            };

//----------------------------------------------------public Functions----------------------------------------------------
//Please use these functions in your sketch

/*----------------------------
Func: DOG-INIT
Desc: Initializes SPI Hardware/Software and DOG Displays
Vars: CS-Pin, MOSI-Pin, SCK-Pin (MOSI=SCK Hardware else Software), A0-Pin (high=data, low=command), p_res = Reset-Pin, type (1=EA DOGS102-6)
------------------------------*/
void dog::initialize(byte p_cs, byte p_si, byte p_clk, byte p_a0, byte p_res, byte type, byte led) 
{
  byte *ptr_init; //pointer to the correct init values
  top_view = false; //default = bottom view
  disp_update = false;  //is true if any function changed the display content. Set to false after data send

  dog::led = led;
  pinMode(led,  OUTPUT);
  TCCR2B = _BV(CS21) | _BV(CS20); //960KHZ PWM
  backlight(0);

  dog::p_a0 = p_a0;
  pinMode(p_a0, OUTPUT);
  spi_initialize(p_cs, p_si, p_clk); //init SPI to Mode 3

  //perform a Reset
  digitalWrite(p_res, LOW);
  pinMode(p_res, OUTPUT);
  delayMicroseconds(10);
  digitalWrite(p_res, HIGH);
  delay(1);

  //Init DOG displays, depending on users choice
  ptr_init = init_DOGS102; //default pointer for wrong parameters
  if(type == DOGS102) 		ptr_init = init_DOGS102;
  
  dog::type = type;

  digitalWrite(p_a0, LOW); //init display
  spi_put(ptr_init, INITLEN);
  
  clear();
}

/*----------------------------
Func: backlight
Desc: sets the LED current / backlight brightness
Vars: brightness
------------------------------*/
void dog::backlight (byte brightness)
{
  if(brightness >= 100)
  {
    brightness = 100;
  }    
  
  analogWrite(led, led_lookup[brightness]);  
}

/*----------------------------
Func: clear
Desc: clears the entire DOG-Display
Vars: ---
------------------------------*/
void dog::clear(void) 
{
  memset(disp_ram, 0, sizeof(disp_ram));
  disp_update = true;	
}

/*----------------------------
Func: clear
Desc: clears rectangle
Vars: two corners of a rectangle
------------------------------*/
void dog::clear (byte xs, byte ys, byte xe, byte ye)
{
  byte page, x, mask, page_cnt;
  
  if(xs > xe)
  {
    x = xs;
    xs = xe;
    xe = x;    
  }
  
  if(xe > XPIXEL-1)
    xe = XPIXEL-1;
  
  if(ys > ye)
  {
    x = ys;
    ys = ye;
    ye = x;    
  }
  if(ye > YPIXEL-1)
    ye = YPIXEL-1;  
  
  page_cnt = ye/8 - ys/8 + 1;
  
  page = ys/8;
  mask = 0xFF >> (8-(ys%8));
  if(page_cnt == 1)
    mask |= 0xFF << (ye%8 + 1);
  

  for(x = xs; x <= xe; x++)
     disp_ram[x + page*XPIXEL] &= mask;


  page_cnt--;
  
  if(page_cnt > 1)
  {
    while(page_cnt-- > 1)
    {
      page++;
      for(x = xs; x <= xe; x++)
        disp_ram[x + page*XPIXEL] = 0;
    }
    page_cnt++;
  }
  
  if(page_cnt)
  {
    page = ye/8;
    mask = 0xFF << (ye%8 + 1);
    
    for(x = xs; x <= xe; x++)
      disp_ram[x + page*XPIXEL] &= mask;
  }
  
  disp_update = true;	
}

/*----------------------------
Func: invert
Desc: inverts rectangle
Vars: two corners of a rectangle
------------------------------*/
void dog::invert (byte xs, byte ys, byte xe, byte ye)
{
  byte page, x, mask, page_cnt;
  
  if(xs > xe)
  {
    x = xs;
    xs = xe;
    xe = x;    
  }
  
  if(xe > XPIXEL-1)
    xe = XPIXEL-1;
  
  if(ys > ye)
  {
    x = ys;
    ys = ye;
    ye = x;    
  }
  if(ye > YPIXEL-1)
    ye = YPIXEL-1;  
  
  page_cnt = ye/8 - ys/8 + 1;
  
  page = ys/8;
  mask = 0xFF >> (8-(ys%8));
  if(page_cnt == 1)
    mask |= 0xFF << (ye%8 + 1);

  mask = ~mask;

  for(x = xs; x <= xe; x++)
     disp_ram[x + page*XPIXEL] ^= mask;


  page_cnt--;
  
  if(page_cnt > 1)
  {
    while(page_cnt-- > 1)
    {
      page++;
      for(x = xs; x <= xe; x++)
        disp_ram[x + page*XPIXEL] ^= 0xFF;
    }
    page_cnt++;
  }
  
  if(page_cnt)
  {
    page = ye/8;
    mask = 0xFF << (ye%8 + 1);
    mask = ~mask;
    
    for(x = xs; x <= xe; x++)
      disp_ram[x + page*XPIXEL] ^= mask;
  }
  
  disp_update = true;  
}

/*----------------------------
Func: contrast
Desc: sets contrast to the DOG-Display
Vars: byte contrast (0..63)
------------------------------*/
void dog::contrast(byte contr) 
{
  command(0x81);   		//double byte command
  command(contr&0x3F);	//contrast has only 6 bits
}

/*----------------------------
Func: view
Desc: ssets the display viewing direction
Vars: direction (top view 0xC8, bottom view (default) = 0xC0)
------------------------------*/
void dog::view(byte direction)  
{
  if(direction == VIEW_TOP)
  {
    top_view = true;
    command(0xA0);
  }
  else
  {
    top_view = false;
    command(0xA1);
  }
  command(direction);
}

/*----------------------------
Func: string
Desc: shows string with selected font on position
Vars: x(0..127/131), y(0..63/127),  font adress in programm memory, stringarray
------------------------------*/
void dog::string(byte x, byte y, const byte *font_adress, const char *str)
{
  unsigned int pos_array; 										//Postion of character data in memory array
  byte x_char, page_cnt, column_cnt, width_max;								//temporary column and page adress, couloumn_cnt tand width_max are used to stay inside display area
  byte start_code, last_code, width, page_height, bytes_p_char;	//font information, needed for calculation
  const char *p_str;
  byte page, shift, pixel_dat;
  
  start_code 	 = pgm_read_byte(&font_adress[2]);  //get first defined character
  last_code	 = pgm_read_byte(&font_adress[3]);  //get last defined character
  width		 = pgm_read_byte(&font_adress[4]);  //width in pixel of one char
  page_height  = pgm_read_byte(&font_adress[6]);  //page count per char
  bytes_p_char = pgm_read_byte(&font_adress[7]);  //bytes per char
  
  page = y/8;
  shift = y%8;

  if(page_height * 8 + y > YPIXEL) //stay inside display area
     page_height = (YPIXEL - y - 1) / 8 + 1;
	
  //The string is displayed character after character. If the font has more then one page, 
  //the top page is printed first, then the next page and so on
  for(page_cnt = 0; page_cnt < page_height; page_cnt++)
  {
    column_cnt = x; //store column for display last column check
    p_str = str;             //temporary pointer to the beginning of the string to print
    while(*p_str != 0)
    {
      if((byte)*p_str < start_code || (byte)*p_str > last_code) //make sure data is valid
        p_str++;
      else
      {
        //calculate positon of ascii character in font array
      	//bytes for header + (ascii - startcode) * bytes per char)
      	pos_array = 8 + (unsigned int)(*p_str++ - start_code) * bytes_p_char;
      	pos_array += page_cnt*width; //get the dot pattern for the part of the char to print
      		
      	if(column_cnt + width > XPIXEL) //stay inside display area
      	  width_max = XPIXEL-column_cnt;
      	else
      	  width_max = width;
      	for(x_char=0; x_char < width_max; x_char++) //print the whole string
      	{
                pixel_dat = pgm_read_byte(&font_adress[pos_array+x_char]);
                disp_ram[(x_char+column_cnt) + ((page_cnt+page)*XPIXEL)]   |= pixel_dat << shift;
                if(page_cnt+page+1 < YPIXEL/8)
                  disp_ram[(x_char+column_cnt) + ((page_cnt+page+1)*XPIXEL)] |= pixel_dat >> (8-shift);
      	}
      	column_cnt += width;
      }
    }
  }
  disp_update = true;
}

/*----------------------------
Func: line
Desc: draws a line
Vars: start/end coordinate
------------------------------*/
/*------------------------------------------------------------------
 Draw line
                            case 3
                              E           
                4 equal  4 s  | 2 steep 2 equal
                      \       |       /
                        \     |     /
                     4 f  \   |   /  2 flat
                            \ | /
          case 5 E------------S------------E case 1
                            / | \
                     6 f  /   |   \  8 f
                        /     |     \
                      /   6 s | 8 s  \
                   6 equal    |         8 equal
                              E           
                            case 7

case 1: horizintal  (= case 5: swap xs <-> xe)
case 7: vartical    (= case 3: swap ys <-> ye)

case 2:             (= case 6: swap xs,xe + ys,ye)
        flat <45     X increment
        steep >=45   Y decrement
        
case 8:             (= case 4:  swap xs,xe + ys,ye)
        flat <45     X increment
        steep >=45   Y decrement
------------------------------------------------------------------*/
void dog::line(byte xs, byte ys, byte xe, byte ye)  
{
  byte temp;
  byte page;
  
  if(xs>XPIXEL-1) //make sure to stay in display area
    xs = XPIXEL-1;
  if(ys>YPIXEL-1)
    ys = YPIXEL-1;
  if(xe>XPIXEL-1)
    xe = XPIXEL-1;
  if(ye>YPIXEL-1)
    ye = YPIXEL-1;
    
  if(ys == ye)      //case 1 and 5
    DrawLine_1_5(xs,xe,ys);
  else if(xs == xe) //case 3 and 7
    DrawLine_3_7(ys,ye,xs);
  else           //all other cases              
  { 
    if(xs > xe)  //make xs smaller then xe, but keep coordinates together
    { 
      temp = xs;             
      xs = xe;
      xe = temp;
      temp = ys;
      ys = ye;
      ye = temp;
     }
    
    temp = xe - xs + 1;   //calculate x distance
    
    if(ys < ye)           //calculate y disatance
      page = ye - ys + 1;
    else                         
      page = ys - ye + 1;
   
    if(temp > page)      //which distance is greater? flat or steep
      DrawLine_2_8_flat(xs,ys, xe,ye);
    else
      DrawLine_2_8_steep(xs,ys, xe,ye);
   }
   disp_update = true;
}

/*----------------------------
Func: rectangle
Desc: shows a pattern filled rectangle on the display
Vars: star/end coordinate and page(0..3/7), thickness of line
------------------------------*/
void dog::rectangle(byte xs, byte ys, byte xe, byte ye)  
{
    line(xs,ys, xe,ys);
    line(xe,ys, xe,ye);
    line(xe,ye, xs,ye);
    line(xs,ye, xs,ys);
}

/*----------------------------
Func: picture
Desc: shows a BLH-picture on the display (see BitMapEdit EA LCD-Tools (http://www.lcd-module.de/support.html))
Vars: x (0..101) and page(0..7), program memory adress of data
------------------------------*/
void dog::picture(byte x, byte y, const byte *pic_adress)  
{
  byte c,p;
  unsigned int line_cnt;
  byte width, page_cnt, pic_width;
  byte page, shift;
  byte pixel_data;

  pic_width = width = pgm_read_byte(&pic_adress[0]);
  if(width + x > XPIXEL) //stay inside display area
    width = XPIXEL - x;
  
  page = y/8;
  shift = y%8;
  if(y+pgm_read_byte(&pic_adress[1]) > YPIXEL)
    page_cnt = (YPIXEL - y - 1) / 8 + 1;
  else
    page_cnt = (pgm_read_byte(&pic_adress[1]) + 7) / 8; //height in pages, add 7 and divide by 8 for getting the used pages (byte boundaries)
  
  line_cnt = 2;
  for(p=0; p<page_cnt; p++)
  {
     for(c=0; c<width; c++)
     {
       pixel_data = pgm_read_byte(&pic_adress[line_cnt + c]);
       disp_ram[(x+c) + ((p+page)*XPIXEL)] |= pixel_data << shift;
       if(p+page+1 < YPIXEL/8)
         disp_ram[(x+c) + ((p+page+1)*XPIXEL)] |= pixel_data >> (8-shift);
     }
     line_cnt += pic_width;
  }
  disp_update = true;
}

/*----------------------------
Func: update
Desc: shows internal display ram content on the screen
Vars: ---
------------------------------*/
void dog::update(void)
{
  byte y,x;
  if(disp_update)
  {
    for(y = 0; y < YPIXEL/8; y++)
    {
      position(0,y);
      digitalWrite(p_cs, LOW);
      digitalWrite(p_a0, HIGH);

      for(x = 0; x < XPIXEL; x++) //clear the whole page line
        spi_out(disp_ram[(x)+(y*XPIXEL)]);
      digitalWrite(p_cs, HIGH);
    }
  }
}

//----------------------------------------------------private Functions----------------------------------------------------
//normally you don't need those functions in your sketch

/*----------------------------
Func: DrawLine_1_5
Desc: horizontal line, more details see line()
Vars: xstart, xend, y
------------------------------*/
void dog::DrawLine_1_5(byte xs, byte xe, byte y)
{
  byte i;
  
  if(xs > xe)    // case 5
  {
    i = xs;       //swap for case 1
    xs = xe;
    xe = i;
  }

  for(i=xs; i <= xe; i++)
    SetPixel(i,y);
}

/*----------------------------
Func: DrawLine_3_7
Desc: vertical line, more details see line()
Vars: ystart, yend, x
------------------------------*/
void dog::DrawLine_3_7(byte ys, byte ye, byte x)
{
  byte i;
  
  if(ys > ye)    // case 3 
  { 
    i = ys;       //swap for case 7
    ys = ye;
    ye = i;
  }
  
  for(i=ys; i <= ye; i++)
    SetPixel(x,i);
}

/*----------------------------
Func: DrawLine_2_8_flat
Desc: 0<line<45, more details see line()
Vars: xs,ys, xe,ye
------------------------------*/
void dog::DrawLine_2_8_flat(byte xs, byte ys, byte xe, byte ye)
{
  byte  y, ya; 
  byte  x, xa, dif;
  char incr;
  
  incr = (ys < ye) ? 1 : -1;
  
  y = ys;
  if(xe > xs)
    xa = xe - xs + 1;
  else
    xa = xs - xe + 1;
  if(ye > ys)
    ya = ye - ys + 1;
  else
    ya = ys - ye + 1;
    
  dif = xa;
  for(x = xs ; x <= xe; x++)
  { 
    SetPixel(x, y);
    if(dif < xa)
    { 
      dif += xa;
      y += incr;
    }
    
    dif -= ya;

  } 
}

/*----------------------------
Func: DrawLine_2_8_steep
Desc: 45<=line<90, more details see line()
Vars: xs,ys, xe,ye
------------------------------*/
void dog::DrawLine_2_8_steep(byte xs, byte ys, byte xe, byte ye)
{
  byte  y, ya; 
  byte  x, xa, dif;
  char incr;
  
  incr = (ys < ye) ? 1 : -1;
  
  x = xs;
  y = ys;
  
  if(xe > xs)
    xa = xe - xs + 1;
  else
    xa = xs - xe + 1;
  if(ye > ys)
    ya = ye - ys + 1;
  else
    ya = ys - ye + 1;
    
  dif = ya;
  for(;;)
   { SetPixel(x, y);
     if(y == ye)
        break;
     y += incr;
     
     if(dif <= xa)
     {
       dif += ya;
        x++;
     }

     dif -= xa;
   }
}

void dog::SetPixel(byte x, byte y)
{
  disp_ram[(x) + ((y/8)*XPIXEL)] |= 0x1 << (y%8);
}


/*----------------------------
Func: position
Desc: sets write pointer in DOG-Display
Vars: column (0..101), page(0..7)
------------------------------*/
void dog::position(byte column, byte page)  
{
  if(top_view)
    column += 30;
  
  command(0x10 + (column>>4)); 	 //MSB adress column
  command(0x00 + (column&0x0F)); //LSB adress column
  command(0xB0 + (page&0x0F)); 	 //adress page	
}

/*----------------------------
Func: command
Desc: Sends a command to the DOG-Display
Vars: data
------------------------------*/
void dog::command(byte dat) 
{
  digitalWrite(p_a0, LOW);
  spi_put_byte(dat);
}

/*----------------------------
Func: data
Desc: Sends data to the DOG-Display
Vars: data
------------------------------*/
void dog::data(byte dat) 
{
  digitalWrite(p_a0, HIGH);
  spi_put_byte(dat);
}

/*----------------------------
Func: spi_initialize
Desc: Initializes SPI Hardware/Software
Vars: CS-Pin, MOSI-Pin, SCK-Pin (MOSI=SCK Hardware else Software)
------------------------------*/
void dog::spi_initialize(byte cs, byte si, byte clk) 
{
  //Set pin Configuration
  p_cs = cs;
  if(si == clk)
  {
    hardware = true;
    p_si = MOSI;
    p_clk = SCK;
  }
  else 
  {
    hardware = false;
    p_si = si;
    p_clk = clk;
  }
	
  // Set CS to deselct slaves
  digitalWrite(p_cs, HIGH);
  pinMode(p_cs, OUTPUT);
	
  // Set Data pin as output
  pinMode(p_si, OUTPUT);

  // Set SPI-Mode 3: CLK idle high, rising edge, MSB first
  digitalWrite(p_clk, HIGH);
  pinMode(p_clk, OUTPUT);
  if(hardware)
  {	
    SPI.begin();
    SPI.setBitOrder(MSBFIRST);
    SPI.setDataMode(SPI_MODE3);
    SPI.setClockDivider(SPI_CLOCK_DIV4);
  }
}

/*----------------------------
Func: spi_put_byte
Desc: Sends one Byte using CS
Vars: data
------------------------------*/
void dog::spi_put_byte(byte dat) 
{
  digitalWrite(p_cs, LOW);
  spi_out(dat);
  digitalWrite(p_cs, HIGH);
}

/*----------------------------
Func: spi_put
Desc: Sends bytes using CS
Vars: ptr to data and len
------------------------------*/
void dog::spi_put(byte *dat, int len) 
{
  digitalWrite(p_cs, LOW);
  do
  {
    spi_out(*dat++);
  }while(--len);
	
  digitalWrite(p_cs, HIGH);
}

/*----------------------------
Func: spi_out
Desc: Sends one Byte, no CS
Vars: data
------------------------------*/
void dog::spi_out(byte dat) 
{
  byte i = 8;
  if(hardware) 
  {
    SPI.transfer(dat);
  }
  else 
  {
    do 
    {
      if(dat & 0x80)
        digitalWrite(p_si, HIGH);
      else
	digitalWrite(p_si, LOW);
      digitalWrite(p_clk, LOW);
      dat <<= 1;
      digitalWrite(p_clk, HIGH);
    }while(--i);
  }
}

